@@ -1,7 +1,7 @@
 /*
  *  dload.c
  *
- *  Copyright (c) 2006-2021 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2024 Pacman Development Team <pacman-dev@lists.archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -28,7 +28,10 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 #include <signal.h>
+#include <dirent.h>
+#include <pwd.h>
 
 #ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h> /* IPPROTO_TCP */
@@ -48,6 +51,73 @@
 #include "log.h"
 #include "util.h"
 #include "handle.h"
+#include "sandbox.h"
+
+
+static const char *get_filename(const char *url)
+{
+	char *filename = strrchr(url, '/');
+	if(filename != NULL) {
+		return filename + 1;
+	}
+
+	/* no slash found, it's a filename */
+	return url;
+}
+
+/* prefix to avoid possible future clash with getumask(3) */
+static mode_t _getumask(void)
+{
+	mode_t mask = umask(0);
+	umask(mask);
+	return mask;
+}
+
+static int finalize_download_file(const char *filename)
+{
+	struct stat st;
+	ASSERT(filename != NULL, return -1);
+	ASSERT(stat(filename, &st) == 0, return -1);
+	if(st.st_size == 0) {
+		unlink(filename);
+                return 1;
+	}
+	ASSERT(chown(filename, 0, 0) != -1, return -1);
+	ASSERT(chmod(filename, ~(_getumask()) & 0666) != -1, return -1);
+	return 0;
+}
+
+static FILE *create_tempfile(struct dload_payload *payload, const char *localpath)
+{
+	int fd;
+	FILE *fp;
+	char *randpath;
+	size_t len;
+
+	/* create a random filename, which is opened with O_EXCL */
+	len = strlen(localpath) + 14 + 1;
+	MALLOC(randpath, len, RET_ERR(payload->handle, ALPM_ERR_MEMORY, NULL));
+	snprintf(randpath, len, "%salpmtmp.XXXXXX", localpath);
+	if((fd = mkstemp(randpath)) == -1 ||
+			fchmod(fd, ~(_getumask()) & 0666) ||
+			!(fp = fdopen(fd, payload->tempfile_openmode))) {
+		unlink(randpath);
+		close(fd);
+		_alpm_log(payload->handle, ALPM_LOG_ERROR,
+				_("failed to create temporary file for download\n"));
+		free(randpath);
+		return NULL;
+	}
+	/* fp now points to our alpmtmp.XXXXXX */
+	free(payload->tempfile_name);
+	payload->tempfile_name = randpath;
+	free(payload->remote_name);
+	STRDUP(payload->remote_name, strrchr(randpath, '/') + 1,
+			fclose(fp); RET_ERR(payload->handle, ALPM_ERR_MEMORY, NULL));
+
+	return fp;
+}
+
 
 #ifdef HAVE_LIBCURL
 
@@ -55,16 +125,16 @@
 #define HOSTNAME_SIZE 256
 
 static int curl_add_payload(alpm_handle_t *handle, CURLM *curlm,
-	struct dload_payload *payload, const char *localpath);
+	struct dload_payload *payload);
 static int curl_gethost(const char *url, char *buffer, size_t buf_len);
 
 /* number of "soft" errors required to blacklist a server, set to 0 to disable
  * server blacklisting */
-const unsigned int server_error_limit = 3;
+const int server_error_limit = 3;
 
 struct server_error_count {
 	char server[HOSTNAME_SIZE];
-	unsigned int errors;
+	int errors;
 };
 
 static struct server_error_count *find_server_errors(alpm_handle_t *handle, const char *server)
@@ -94,22 +164,34 @@ static struct server_error_count *find_server_errors(alpm_handle_t *handle, cons
 	}
 }
 
+/* skip for hard errors or too many soft errors */
 static int should_skip_server(alpm_handle_t *handle, const char *server)
 {
 	struct server_error_count *h;
 	if(server_error_limit && (h = find_server_errors(handle, server)) ) {
-		return h->errors >= server_error_limit;
+		return h->errors < 0 || h->errors >= server_error_limit;
 	}
 	return 0;
 }
 
-static void server_increment_error(alpm_handle_t *handle, const char *server, int count)
+/* only skip for hard errors */
+static int should_skip_cache_server(alpm_handle_t *handle, const char *server)
+{
+	struct server_error_count *h;
+	if(server_error_limit && (h = find_server_errors(handle, server)) ) {
+		return h->errors < 0;
+	}
+	return 0;
+}
+
+/* block normal servers after too many errors */
+static void server_soft_error(alpm_handle_t *handle, const char *server)
 {
 	struct server_error_count *h;
 	if(server_error_limit
 			&& (h = find_server_errors(handle, server))
 			&& !should_skip_server(handle, server) ) {
-		h->errors += count;
+		h->errors++;
 
 		if(should_skip_server(handle, server)) {
 			_alpm_log(handle, ALPM_LOG_WARNING,
@@ -119,42 +201,50 @@ static void server_increment_error(alpm_handle_t *handle, const char *server, in
 	}
 }
 
-static void server_soft_error(alpm_handle_t *handle, const char *server)
-{
-	server_increment_error(handle, server, 1);
-}
-
+/* immediate block for both servers and cache servers */
 static void server_hard_error(alpm_handle_t *handle, const char *server)
 {
-	server_increment_error(handle, server, server_error_limit);
-}
+	struct server_error_count *h;
+	if(server_error_limit && (h = find_server_errors(handle, server))) {
+		if(h->errors != -1) {
+			/* always set even if already skipped for soft errors
+			 * to disable cache servers too */
+			h->errors = -1;
 
-static const char *get_filename(const char *url)
-{
-	char *filename = strrchr(url, '/');
-	if(filename != NULL) {
-		return filename + 1;
+			_alpm_log(handle, ALPM_LOG_WARNING,
+					_("fatal error from %s, skipping for the remainder of this transaction\n"),
+					h->server);
+		}
 	}
-
-	/* no slash found, it's a filename */
-	return url;
 }
 
-static char *get_fullpath(const char *path, const char *filename,
-		const char *suffix)
+static const char *payload_next_server(struct dload_payload *payload)
 {
-	char *filepath;
-	/* len = localpath len + filename len + suffix len + null */
-	size_t len = strlen(path) + strlen(filename) + strlen(suffix) + 1;
-	MALLOC(filepath, len, return NULL);
-	snprintf(filepath, len, "%s%s%s", path, filename, suffix);
-
-	return filepath;
+	while(payload->cache_servers
+			&& should_skip_cache_server(payload->handle, payload->cache_servers->data)) {
+		payload->cache_servers = payload->cache_servers->next;
+	}
+	if(payload->cache_servers) {
+		const char *server = payload->cache_servers->data;
+		payload->cache_servers = payload->cache_servers->next;
+		payload->request_errors_ok = 1;
+		return server;
+	}
+	while(payload->servers
+			&& should_skip_server(payload->handle, payload->servers->data)) {
+		payload->servers = payload->servers->next;
+	}
+	if(payload->servers) {
+		const char *server = payload->servers->data;
+		payload->servers = payload->servers->next;
+		payload->request_errors_ok = payload->errors_ok;
+		return server;
+	}
+	return NULL;
 }
 
 enum {
-	ABORT_SIGINT = 1,
-	ABORT_OVER_MAXFILESIZE
+	ABORT_OVER_MAXFILESIZE = 1,
 };
 
 static int dload_interrupted;
@@ -263,45 +353,12 @@ static int utimes_long(const char *path, long seconds)
 	return 0;
 }
 
-/* prefix to avoid possible future clash with getumask(3) */
-static mode_t _getumask(void)
-{
-	mode_t mask = umask(0);
-	umask(mask);
-	return mask;
-}
-
 static size_t dload_parseheader_cb(void *ptr, size_t size, size_t nmemb, void *user)
 {
 	size_t realsize = size * nmemb;
-	const char *fptr, *endptr = NULL;
-	const char * const cd_header = "Content-Disposition:";
-	const char * const fn_key = "filename=";
 	struct dload_payload *payload = (struct dload_payload *)user;
 	long respcode;
-
-	if(_alpm_raw_ncmp(cd_header, ptr, strlen(cd_header)) == 0) {
-		if((fptr = strstr(ptr, fn_key))) {
-			fptr += strlen(fn_key);
-
-			/* find the end of the field, which is either a semi-colon, or the end of
-			 * the data. As per curl_easy_setopt(3), we cannot count on headers being
-			 * null terminated, so we look for the closing \r\n */
-			endptr = fptr + strcspn(fptr, ";\r\n") - 1;
-
-			/* remove quotes */
-			if(*fptr == '"' && *endptr == '"') {
-				fptr++;
-				endptr--;
-			}
-
-			/* avoid information leakage with badly formed headers */
-			if(endptr > fptr) {
-				STRNDUP(payload->content_disp_name, fptr, endptr - fptr + 1,
-						RET_ERR(payload->handle, ALPM_ERR_MEMORY, realsize));
-			}
-		}
-	}
+	(void) ptr;
 
 	curl_easy_getinfo(payload->curl, CURLINFO_RESPONSE_CODE, &respcode);
 	if(payload->respcode != respcode) {
@@ -356,14 +413,13 @@ static void curl_set_handle_opts(CURL *curl, struct dload_payload *payload)
 		curl_easy_setopt(curl, CURLOPT_USERAGENT, useragent);
 	}
 
-	if(!payload->force && payload->destfile_name &&
-			stat(payload->destfile_name, &st) == 0) {
+	if(!payload->force && payload->mtime_existing_file) {
 		/* start from scratch, but only download if our local is out of date. */
 		curl_easy_setopt(curl, CURLOPT_TIMECONDITION, CURL_TIMECOND_IFMODSINCE);
-		curl_easy_setopt(curl, CURLOPT_TIMEVALUE, (long)st.st_mtime);
+		curl_easy_setopt(curl, CURLOPT_TIMEVALUE, payload->mtime_existing_file);
 		_alpm_log(handle, ALPM_LOG_DEBUG,
 				"%s: using time condition %ld\n",
-				payload->remote_name, (long)st.st_mtime);
+				payload->remote_name, (long)payload->mtime_existing_file);
 	} else if(stat(payload->tempfile_name, &st) == 0 && payload->allow_resume) {
 		/* a previous partial download exists, resume from end of file. */
 		payload->tempfile_openmode = "ab";
@@ -375,62 +431,28 @@ static void curl_set_handle_opts(CURL *curl, struct dload_payload *payload)
 	}
 }
 
-static FILE *create_tempfile(struct dload_payload *payload, const char *localpath)
-{
-	int fd;
-	FILE *fp;
-	char *randpath;
-	size_t len;
-
-	/* create a random filename, which is opened with O_EXCL */
-	len = strlen(localpath) + 14 + 1;
-	MALLOC(randpath, len, RET_ERR(payload->handle, ALPM_ERR_MEMORY, NULL));
-	snprintf(randpath, len, "%salpmtmp.XXXXXX", localpath);
-	if((fd = mkstemp(randpath)) == -1 ||
-			fchmod(fd, ~(_getumask()) & 0666) ||
-			!(fp = fdopen(fd, payload->tempfile_openmode))) {
-		unlink(randpath);
-		close(fd);
-		_alpm_log(payload->handle, ALPM_LOG_ERROR,
-				_("failed to create temporary file for download\n"));
-		free(randpath);
-		return NULL;
-	}
-	/* fp now points to our alpmtmp.XXXXXX */
-	free(payload->tempfile_name);
-	payload->tempfile_name = randpath;
-	free(payload->remote_name);
-	STRDUP(payload->remote_name, strrchr(randpath, '/') + 1,
-			fclose(fp); RET_ERR(payload->handle, ALPM_ERR_MEMORY, NULL));
-
-	return fp;
-}
-
 /* Return 0 if retry was successful, -1 otherwise */
 static int curl_retry_next_server(CURLM *curlm, CURL *curl, struct dload_payload *payload)
 {
-	const char *server;
+	const char *server = NULL;
 	size_t len;
 	struct stat st;
 	alpm_handle_t *handle = payload->handle;
 
-	while(payload->servers && should_skip_server(handle, payload->servers->data)) {
-		payload->servers = payload->servers->next;
-	}
-	if(!payload->servers) {
+	if((server = payload_next_server(payload)) == NULL) {
 		_alpm_log(payload->handle, ALPM_LOG_DEBUG,
 				"%s: no more servers to retry\n", payload->remote_name);
 		return -1;
 	}
-	server = payload->servers->data;
-	payload->servers = payload->servers->next;
 
 	/* regenerate a new fileurl */
 	FREE(payload->fileurl);
 	len = strlen(server) + strlen(payload->filepath) + 2;
 	MALLOC(payload->fileurl, len, RET_ERR(handle, ALPM_ERR_MEMORY, -1));
 	snprintf(payload->fileurl, len, "%s/%s", server, payload->filepath);
-
+	_alpm_log(handle, ALPM_LOG_DEBUG,
+			"%s: retrying from %s\n",
+			payload->remote_name, payload->fileurl);
 
 	fflush(payload->localf);
 
@@ -471,10 +493,9 @@ static int curl_retry_next_server(CURLM *curlm, CURL *curl, struct dload_payload
  * Returns -1 if an error happened for a required file
  * Returns -2 if an error happened for an optional file
  */
-static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
-		const char *localpath, int *active_downloads_num)
+static int curl_check_finished_download(alpm_handle_t *handle, CURLM *curlm, CURLMsg *msg,
+		int *active_downloads_num)
 {
-	alpm_handle_t *handle = NULL;
 	struct dload_payload *payload = NULL;
 	CURL *curl = msg->easy_handle;
 	CURLcode curlerr;
@@ -489,12 +510,11 @@ static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
 
 	curlerr = curl_easy_getinfo(curl, CURLINFO_PRIVATE, &payload);
 	ASSERT(curlerr == CURLE_OK, RET_ERR(handle, ALPM_ERR_LIBCURL, -1));
-	handle = payload->handle;
 
 	curl_gethost(payload->fileurl, hostname, sizeof(hostname));
 	curlerr = msg->data.result;
-	_alpm_log(handle, ALPM_LOG_DEBUG, "%s: curl returned result %d from transfer\n",
-			payload->remote_name, curlerr);
+	_alpm_log(handle, ALPM_LOG_DEBUG, "%s: %s returned result %d from transfer\n",
+			payload->remote_name, "curl", curlerr);
 
 	/* was it a success? */
 	switch(curlerr) {
@@ -503,7 +523,7 @@ static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
 			_alpm_log(handle, ALPM_LOG_DEBUG, "%s: response code %ld\n",
 					payload->remote_name, payload->respcode);
 			if(payload->respcode >= 400) {
-				if(!payload->errors_ok) {
+				if(!payload->request_errors_ok) {
 					handle->pm_errno = ALPM_ERR_RETRIEVE;
 					/* non-translated message is same as libcurl */
 					snprintf(payload->error_buffer, sizeof(payload->error_buffer),
@@ -554,11 +574,10 @@ static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
 				(*active_downloads_num)++;
 				return 2;
 			} else {
-				payload->unlink_on_fail = 1;
 				goto cleanup;
 			}
 		default:
-			if(!payload->errors_ok) {
+			if(!payload->request_errors_ok) {
 				handle->pm_errno = ALPM_ERR_LIBCURL;
 				_alpm_log(handle, ALPM_LOG_ERROR,
 						_("failed retrieving file '%s' from %s : %s\n"),
@@ -588,31 +607,6 @@ static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
 	curl_easy_getinfo(curl, CURLINFO_CONDITION_UNMET, &timecond);
 	curl_easy_getinfo(curl, CURLINFO_EFFECTIVE_URL, &effective_url);
 
-	if(payload->trust_remote_name) {
-		if(payload->content_disp_name) {
-			/* content-disposition header has a better name for our file */
-			free(payload->destfile_name);
-			payload->destfile_name = get_fullpath(localpath,
-				get_filename(payload->content_disp_name), "");
-		} else {
-			const char *effective_filename = strrchr(effective_url, '/');
-
-			if(effective_filename && strlen(effective_filename) > 2) {
-				effective_filename++;
-
-				/* if destfile was never set, we wrote to a tempfile. even if destfile is
-				 * set, we may have followed some redirects and the effective url may
-				 * have a better suggestion as to what to name our file. in either case,
-				 * refactor destfile to this newly derived name. */
-				if(!payload->destfile_name || strcmp(effective_filename,
-							strrchr(payload->destfile_name, '/') + 1) != 0) {
-					free(payload->destfile_name);
-					payload->destfile_name = get_fullpath(localpath, effective_filename, "");
-				}
-			}
-		}
-	}
-
 	/* Let's check if client requested downloading accompanion *.sig file */
 	if(!payload->signature && payload->download_signature && curlerr == CURLE_OK && payload->respcode < 400) {
 		struct dload_payload *sig = NULL;
@@ -644,22 +638,23 @@ static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
 		MALLOC(sig->fileurl, len, FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
 		snprintf(sig->fileurl, len, "%s.sig", url);
 
-		if(payload->trust_remote_name) {
-			/* In this case server might provide a new name for the main payload.
-			 * Choose *.sig filename based on this new name.
-			 */
-			const char *final_file = get_filename(realname);
-			int remote_name_len = strlen(final_file) + 5;
-			MALLOC(sig->remote_name, remote_name_len, FREE(sig->fileurl); FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
-			snprintf(sig->remote_name, remote_name_len, "%s.sig", final_file);
-		}
+		int remote_name_len = strlen(payload->remote_name) + 5;
+		MALLOC(sig->remote_name, remote_name_len, _alpm_dload_payload_reset(sig);
+			FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
+		snprintf(sig->remote_name, remote_name_len, "%s.sig", payload->remote_name);
 
 		/* force the filename to be realname + ".sig" */
 		int destfile_name_len = strlen(realname) + 5;
-		MALLOC(sig->destfile_name, destfile_name_len, FREE(sig->remote_name);
-				FREE(sig->fileurl); FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
+		MALLOC(sig->destfile_name, destfile_name_len, _alpm_dload_payload_reset(sig);
+				FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
 		snprintf(sig->destfile_name, destfile_name_len, "%s.sig", realname);
 
+		int tempfile_name_len = strlen(realname) + 10;
+		MALLOC(sig->tempfile_name, tempfile_name_len, _alpm_dload_payload_reset(sig);
+				FREE(sig); GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
+		snprintf(sig->tempfile_name, tempfile_name_len, "%s.sig.part", realname);
+
+
 		sig->signature = 1;
 		sig->handle = handle;
 		sig->force = payload->force;
@@ -668,7 +663,7 @@ static int curl_check_finished_download(CURLM *curlm, CURLMsg *msg,
 		/* set hard upper limit of 16KiB */
 		sig->max_size = 16 * 1024;
 
-		curl_add_payload(handle, curlm, sig, localpath);
+		curl_add_payload(handle, curlm, sig);
 		(*active_downloads_num)++;
 	}
 
@@ -751,7 +746,7 @@ cleanup:
  * Returns -1 if am error happened while starting a new download
  */
 static int curl_add_payload(alpm_handle_t *handle, CURLM *curlm,
-		struct dload_payload *payload, const char *localpath)
+		struct dload_payload *payload)
 {
 	size_t len;
 	CURL *curl = NULL;
@@ -764,53 +759,24 @@ static int curl_add_payload(alpm_handle_t *handle, CURLM *curlm,
 	if(payload->fileurl) {
 		ASSERT(!payload->servers, GOTO_ERR(handle, ALPM_ERR_WRONG_ARGS, cleanup));
 		ASSERT(!payload->filepath, GOTO_ERR(handle, ALPM_ERR_WRONG_ARGS, cleanup));
+		payload->request_errors_ok = payload->errors_ok;
 	} else {
-		const char *server;
-		while(payload->servers && should_skip_server(handle, payload->servers->data)) {
-			payload->servers = payload->servers->next;
-		}
+		const char *server = payload_next_server(payload);
 
-		ASSERT(payload->servers, GOTO_ERR(handle, ALPM_ERR_SERVER_NONE, cleanup));
+		ASSERT(server, GOTO_ERR(handle, ALPM_ERR_SERVER_NONE, cleanup));
 		ASSERT(payload->filepath, GOTO_ERR(handle, ALPM_ERR_WRONG_ARGS, cleanup));
 
-		server = payload->servers->data;
-		payload->servers = payload->servers->next;
-
 		len = strlen(server) + strlen(payload->filepath) + 2;
 		MALLOC(payload->fileurl, len, GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
 		snprintf(payload->fileurl, len, "%s/%s", server, payload->filepath);
 	}
 
 	payload->tempfile_openmode = "wb";
-	if(!payload->remote_name) {
-		STRDUP(payload->remote_name, get_filename(payload->fileurl),
-			GOTO_ERR(handle, ALPM_ERR_MEMORY, cleanup));
-	}
 	if(curl_gethost(payload->fileurl, hostname, sizeof(hostname)) != 0) {
 		_alpm_log(handle, ALPM_LOG_ERROR, _("url '%s' is invalid\n"), payload->fileurl);
 		GOTO_ERR(handle, ALPM_ERR_SERVER_BAD_URL, cleanup);
 	}
 
-	if(!payload->random_partfile && payload->remote_name && strlen(payload->remote_name) > 0) {
-		if(!payload->destfile_name) {
-			payload->destfile_name = get_fullpath(localpath, payload->remote_name, "");
-		}
-		payload->tempfile_name = get_fullpath(localpath, payload->remote_name, ".part");
-		if(!payload->destfile_name || !payload->tempfile_name) {
-			goto cleanup;
-		}
-	} else {
-		/* We want a random filename or the URL does not contain a filename, so download to a
-		 * temporary location. We can not support resuming this kind of download; any partial
-		 * transfers will be destroyed */
-		payload->unlink_on_fail = 1;
-
-		payload->localf = create_tempfile(payload, localpath);
-		if(payload->localf == NULL) {
-			goto cleanup;
-		}
-	}
-
 	curl_set_handle_opts(curl, payload);
 
 	if(payload->max_size == payload->initial_size && payload->max_size != 0) {
@@ -851,7 +817,7 @@ cleanup:
 }
 
 /*
- * Use to sort payloads by max size in decending order (largest -> smallest)
+ * Use to sort payloads by max size in descending order (largest -> smallest)
  */
 static int compare_dload_payload_sizes(const void *left_ptr, const void *right_ptr)
 {
@@ -868,8 +834,7 @@ static int compare_dload_payload_sizes(const void *left_ptr, const void *right_p
  * Returns 1 if no files were downloaded and all errors were non-fatal
  */
 static int curl_download_internal(alpm_handle_t *handle,
-		alpm_list_t *payloads /* struct dload_payload */,
-		const char *localpath)
+		alpm_list_t *payloads /* struct dload_payload */)
 {
 	int active_downloads_num = 0;
 	int err = 0;
@@ -877,24 +842,27 @@ static int curl_download_internal(alpm_handle_t *handle,
 	int updated = 0; /* was a file actually updated */
 	CURLM *curlm = handle->curlm;
 	size_t payloads_size = alpm_list_count(payloads);
+	alpm_list_t *p;
 
 	/* Sort payloads by package size */
+	payloads = alpm_list_copy(payloads);
 	payloads = alpm_list_msort(payloads, payloads_size, &compare_dload_payload_sizes);
+	p = payloads;
 
-	while(active_downloads_num > 0 || payloads) {
+	while(active_downloads_num > 0 || p) {
 		CURLMcode mc;
 
-		for(; active_downloads_num < max_streams && payloads; active_downloads_num++) {
-			struct dload_payload *payload = payloads->data;
+		for(; active_downloads_num < max_streams && p; active_downloads_num++) {
+			struct dload_payload *payload = p->data;
 
-			if(curl_add_payload(handle, curlm, payload, localpath) == 0) {
-				payloads = payloads->next;
+			if(curl_add_payload(handle, curlm, payload) == 0) {
+				p = p->next;
 			} else {
 				/* The payload failed to start. Do not start any new downloads.
 				 * Wait until all active downloads complete.
 				 */
 				_alpm_log(handle, ALPM_LOG_ERROR, _("failed to setup a download payload for %s\n"), payload->remote_name);
-				payloads = NULL;
+				p = NULL;
 				err = -1;
 			}
 		}
@@ -906,7 +874,7 @@ static int curl_download_internal(alpm_handle_t *handle,
 
 		if(mc != CURLM_OK) {
 			_alpm_log(handle, ALPM_LOG_ERROR, _("curl returned error %d from transfer\n"), mc);
-			payloads = NULL;
+			p = NULL;
 			err = -1;
 		}
 
@@ -917,13 +885,13 @@ static int curl_download_internal(alpm_handle_t *handle,
 				break;
 			}
 			if(msg->msg == CURLMSG_DONE) {
-				int ret = curl_check_finished_download(curlm, msg,
-						localpath, &active_downloads_num);
+				int ret = curl_check_finished_download(handle, curlm, msg,
+						&active_downloads_num);
 				if(ret == -1) {
 					/* if current payload failed to download then stop adding new payloads but wait for the
 					 * current ones
 					 */
-					payloads = NULL;
+					p = NULL;
 					err = -1;
 				} else if(ret == 0) {
 					updated = 1;
@@ -934,23 +902,297 @@ static int curl_download_internal(alpm_handle_t *handle,
 		}
 	}
 
-	_alpm_log(handle, ALPM_LOG_DEBUG, "curl_download_internal return code is %d\n", err);
-	return err ? -1 : updated ? 0 : 1;
+	int ret = err ? -1 : updated ? 0 : 1;
+	_alpm_log(handle, ALPM_LOG_DEBUG, "curl_download_internal return code is %d\n", ret);
+	alpm_list_free(payloads);
+	return ret;
+}
+
+/* Download the requested files by launching a process inside a sandbox.
+ * Returns -1 if an error happened for a required file
+ * Returns 0 if a payload was actually downloaded
+ * Returns 1 if no files were downloaded and all errors were non-fatal
+ */
+static int curl_download_internal_sandboxed(alpm_handle_t *handle,
+		alpm_list_t *payloads /* struct dload_payload */,
+		const char *localpath)
+{
+	int pid, err = 0, ret = -1, callbacks_fd[2];
+	sigset_t oldblock;
+	struct sigaction sa_ign, oldint, oldquit;
+	_alpm_sandbox_callback_context callbacks_ctx;
+
+	sigemptyset(&sa_ign.sa_mask);
+	sa_ign.sa_handler = SIG_IGN;
+	sa_ign.sa_flags=0;
+
+	if(pipe(callbacks_fd) != 0) {
+		return -1;
+	}
+
+	sigaction(SIGINT, &sa_ign, &oldint);
+	sigaction(SIGQUIT, &sa_ign, &oldquit);
+	sigaddset(&sa_ign.sa_mask, SIGCHLD);
+	sigprocmask(SIG_BLOCK, &sa_ign.sa_mask, &oldblock);
+
+	pid = fork();
+	if(pid == -1) {
+		/* fork failed, make sure errno is preserved after cleanup */
+		err = errno;
+	}
+
+	/* child */
+	if(pid == 0) {
+		close(callbacks_fd[0]);
+		fcntl(callbacks_fd[1], F_SETFD, FD_CLOEXEC);
+		callbacks_ctx.callback_pipe = callbacks_fd[1];
+		alpm_option_set_logcb(handle, _alpm_sandbox_cb_log, &callbacks_ctx);
+		alpm_option_set_dlcb(handle, _alpm_sandbox_cb_dl, &callbacks_ctx);
+		alpm_option_set_fetchcb(handle, NULL, NULL);
+		alpm_option_set_eventcb(handle, NULL, NULL);
+		alpm_option_set_questioncb(handle, NULL, NULL);
+		alpm_option_set_progresscb(handle, NULL, NULL);
+
+		/* restore default signal handling in the child */
+		_alpm_reset_signals();
+
+		/* cwd to the download directory */
+		ret = chdir(localpath);
+		if(ret != 0) {
+			handle->pm_errno = ALPM_ERR_NOT_A_DIR;
+			_alpm_log(handle, ALPM_LOG_ERROR, _("could not chdir to download directory %s\n"), localpath);
+			ret = -1;
+		} else {
+			ret = alpm_sandbox_setup_child(handle, handle->sandboxuser, localpath);
+			if (ret != 0) {
+				_alpm_log(handle, ALPM_LOG_ERROR, _("switching to sandbox user '%s' failed!\n"), handle->sandboxuser);
+				_Exit(2);
+			}
+
+			ret = curl_download_internal(handle, payloads);
+		}
+
+		/* pass the result back to the parent */
+		if(ret == 0) {
+			/* a payload was actually downloaded */
+			_Exit(0);
+		}
+		else if(ret == 1) {
+			/* no files were downloaded and all errors were non-fatal */
+			_Exit(1);
+		}
+		else {
+			/* an error happened for a required file */
+			_Exit(2);
+		}
+	}
+
+	/* parent */
+	close(callbacks_fd[1]);
+
+	if(pid != -1)  {
+		bool had_error = false;
+		while(true) {
+			_alpm_sandbox_callback_t callback_type;
+			ssize_t got = read(callbacks_fd[0], &callback_type, sizeof(callback_type));
+			if(got < 0 || (size_t)got != sizeof(callback_type)) {
+				had_error = true;
+				break;
+			}
+
+			if(callback_type == ALPM_SANDBOX_CB_DOWNLOAD) {
+				if(!_alpm_sandbox_process_cb_download(handle, callbacks_fd[0])) {
+					had_error = true;
+					break;
+				}
+			}
+			else if(callback_type == ALPM_SANDBOX_CB_LOG) {
+				if(!_alpm_sandbox_process_cb_log(handle, callbacks_fd[0])) {
+					had_error = true;
+					break;
+				}
+			}
+		}
+
+
+		if(had_error) {
+			kill(pid, SIGTERM);
+		}
+
+		int wret;
+		while((wret = waitpid(pid, &ret, 0)) == -1 && errno == EINTR);
+		if(wret > 0) {
+			if(!WIFEXITED(ret)) {
+				/* the child did not terminate normally */
+				ret = -1;
+			}
+			else {
+				ret = WEXITSTATUS(ret);
+				if(ret != 0) {
+					if(ret == 2) {
+						/* an error happened for a required file, or unexpected exit status */
+						handle->pm_errno = ALPM_ERR_RETRIEVE;
+						ret = -1;
+					}
+					else {
+						handle->pm_errno = ALPM_ERR_RETRIEVE;
+						ret = 1;
+					}
+				}
+			}
+		}
+		else {
+			/* waitpid failed */
+			err = errno;
+		}
+	}
+
+	close(callbacks_fd[0]);
+
+	sigaction(SIGINT, &oldint, NULL);
+	sigaction(SIGQUIT, &oldquit, NULL);
+	sigprocmask(SIG_SETMASK, &oldblock, NULL);
+
+	if(err) {
+		errno = err;
+		ret = -1;
+	}
+	return ret;
 }
 
 #endif
 
+static int payload_download_fetchcb(struct dload_payload *payload,
+		const char *server, const char *localpath)
+{
+	int ret;
+	char *fileurl;
+	alpm_handle_t *handle = payload->handle;
+
+	size_t len = strlen(server) + strlen(payload->filepath) + 2;
+	MALLOC(fileurl, len, RET_ERR(handle, ALPM_ERR_MEMORY, -1));
+	snprintf(fileurl, len, "%s/%s", server, payload->filepath);
+
+	ret = handle->fetchcb(handle->fetchcb_ctx, fileurl, localpath, payload->force);
+	free(fileurl);
+
+	return ret;
+}
+
+static int move_file(const char *filepath, const char *directory)
+{
+	ASSERT(filepath != NULL, return -1);
+	ASSERT(directory != NULL, return -1);
+	int ret = finalize_download_file(filepath);
+	if(ret != 0) {
+		return ret;
+	}
+	const char *filename = mbasename(filepath);
+	char *dest = _alpm_get_fullpath(directory, filename, "");
+	if(rename(filepath, dest)) {
+		FREE(dest);
+		return -1;
+	}
+	FREE(dest);
+	return 0;
+}
+
+static int finalize_download_locations(alpm_list_t *payloads, const char *localpath)
+{
+	ASSERT(payloads != NULL, return -1);
+	ASSERT(localpath != NULL, return -1);
+	alpm_list_t *p;
+	int returnvalue = 0;
+	for(p = payloads; p; p = p->next) {
+		struct dload_payload *payload = p->data;
+		if(payload->tempfile_name) {
+			move_file(payload->tempfile_name, localpath);
+		}
+		if(payload->destfile_name) {
+			int ret = move_file(payload->destfile_name, localpath);
+
+			if(ret == -1) {
+				/* ignore error if the file already existed - only signature file was downloaded */
+				if(payload->mtime_existing_file == 0) {
+					returnvalue = -1;
+				}
+			}
+
+			if (payload->download_signature) {
+				const char sig_suffix[] = ".sig";
+				char *sig_filename = NULL;
+				size_t sig_filename_len = strlen(payload->destfile_name) + sizeof(sig_suffix);
+				MALLOC(sig_filename, sig_filename_len, continue);
+				snprintf(sig_filename, sig_filename_len, "%s%s", payload->destfile_name, sig_suffix);
+				move_file(sig_filename, localpath);
+				FREE(sig_filename);
+			}
+		}
+	}
+	return returnvalue;
+}
+
+static void prepare_resumable_downloads(alpm_list_t *payloads, const char *localpath,
+		const char *user)
+{
+	struct passwd const *pw = NULL;
+	ASSERT(payloads != NULL, return);
+	ASSERT(localpath != NULL, return);
+	if(user != NULL) {
+		ASSERT((pw = getpwnam(user)) != NULL, return);
+	}
+	alpm_list_t *p;
+	for(p = payloads; p; p = p->next) {
+		struct dload_payload *payload = p->data;
+		if(payload->destfile_name) {
+			const char *destfilename = mbasename(payload->destfile_name);
+			char *dest = _alpm_get_fullpath(localpath, destfilename, "");
+			struct stat deststat;
+			if(stat(dest, &deststat) == 0 && deststat.st_size != 0) {
+				payload->mtime_existing_file = deststat.st_mtime;
+			}
+			FREE(dest);
+		}
+		if(!payload->tempfile_name) {
+			continue;
+		}
+		const char *filename = mbasename(payload->tempfile_name);
+		char *src = _alpm_get_fullpath(localpath, filename, "");
+		struct stat st;
+		if(stat(src, &st) != 0 || st.st_size == 0) {
+			FREE(src);
+			continue;
+		}
+		if(rename(src, payload->tempfile_name) != 0) {
+			FREE(src);
+			continue;
+		}
+		if(pw != NULL) {
+			ASSERT(chown(payload->tempfile_name, pw->pw_uid, pw->pw_gid), return);
+		}
+		FREE(src);
+	}
+}
+
 /* Returns -1 if an error happened for a required file
  * Returns 0 if a payload was actually downloaded
  * Returns 1 if no files were downloaded and all errors were non-fatal
  */
 int _alpm_download(alpm_handle_t *handle,
 		alpm_list_t *payloads /* struct dload_payload */,
-		const char *localpath)
+		const char *localpath,
+		const char *temporary_localpath)
 {
+	int ret;
+	prepare_resumable_downloads(payloads, localpath, handle->sandboxuser);
+
 	if(handle->fetchcb == NULL) {
 #ifdef HAVE_LIBCURL
-		return curl_download_internal(handle, payloads, localpath);
+		if(handle->sandboxuser) {
+			ret = curl_download_internal_sandboxed(handle, payloads, temporary_localpath);
+		} else {
+			ret = curl_download_internal(handle, payloads);
+		}
 #else
 		RET_ERR(handle, ALPM_ERR_EXTERNAL_DOWNLOAD, -1);
 #endif
@@ -960,21 +1202,56 @@ int _alpm_download(alpm_handle_t *handle,
 		for(p = payloads; p; p = p->next) {
 			struct dload_payload *payload = p->data;
 			alpm_list_t *s;
-			int ret = -1;
+			ret = -1;
 
 			if(payload->fileurl) {
-				ret = handle->fetchcb(handle->fetchcb_ctx, payload->fileurl, localpath, payload->force);
+				ret = handle->fetchcb(handle->fetchcb_ctx, payload->fileurl, temporary_localpath, payload->force);
+				if (ret != -1 && payload->download_signature) {
+					/* Download signature if requested */
+					char *sig_fileurl;
+					size_t sig_len = strlen(payload->fileurl) + 5;
+					int retsig = -1;
+
+					MALLOC(sig_fileurl, sig_len, RET_ERR(handle, ALPM_ERR_MEMORY, -1));
+					snprintf(sig_fileurl, sig_len, "%s.sig", payload->fileurl);
+
+					retsig = handle->fetchcb(handle->fetchcb_ctx, sig_fileurl, temporary_localpath,  payload->force);
+					free(sig_fileurl);
+
+					if(!payload->signature_optional) {
+						ret = retsig;
+					}
+				}
 			} else {
-				for(s = payload->servers; s && ret == -1; s = s->next) {
-					const char *server = s->data;
-					char *fileurl;
+				for(s = payload->cache_servers; s; s = s->next) {
+					ret = payload_download_fetchcb(payload, s->data, temporary_localpath);
+					if (ret != -1) {
+						goto download_signature;
+					}
+				}
+				for(s = payload->servers; s; s = s->next) {
+					ret = payload_download_fetchcb(payload, s->data, temporary_localpath);
+					if (ret != -1) {
+						goto download_signature;
+					}
+				}
 
-					size_t len = strlen(server) + strlen(payload->filepath) + 2;
-					MALLOC(fileurl, len, RET_ERR(handle, ALPM_ERR_MEMORY, -1));
-					snprintf(fileurl, len, "%s/%s", server, payload->filepath);
+download_signature:
+				if (ret != -1 && payload->download_signature) {
+					/* Download signature if requested */
+					char *sig_fileurl;
+					size_t sig_len = strlen(s->data) + strlen(payload->filepath) + 6;
+					int retsig = -1;
 
-					ret = handle->fetchcb(handle->fetchcb_ctx, fileurl, localpath, payload->force);
-					free(fileurl);
+					MALLOC(sig_fileurl, sig_len, RET_ERR(handle, ALPM_ERR_MEMORY, -1));
+					snprintf(sig_fileurl, sig_len, "%s/%s.sig", (const char *)(s->data), payload->filepath);
+
+					retsig = handle->fetchcb(handle->fetchcb_ctx, sig_fileurl, temporary_localpath, payload->force);
+					free(sig_fileurl);
+
+					if(!payload->signature_optional) {
+						ret = retsig;
+					}
 				}
 			}
 
@@ -984,8 +1261,13 @@ int _alpm_download(alpm_handle_t *handle,
 				updated = 1;
 			}
 		}
-		return updated ? 0 : 1;
+		ret = updated ? 0 : 1;
 	}
+
+	if (finalize_download_locations(payloads, localpath) != 0 && ret == 0) {
+		return -1;
+	}
+	return ret;
 }
 
 static char *filecache_find_url(alpm_handle_t *handle, const char *url)
@@ -1008,6 +1290,7 @@ int SYMEXPORT alpm_fetch_pkgurl(alpm_handle_t *handle, const alpm_list_t *urls,
 	  alpm_list_t **fetched)
 {
 	const char *cachedir;
+	char *temporary_cachedir = NULL;
 	alpm_list_t *payloads = NULL;
 	const alpm_list_t *i;
 	alpm_event_t event = {0};
@@ -1017,6 +1300,8 @@ int SYMEXPORT alpm_fetch_pkgurl(alpm_handle_t *handle, const alpm_list_t *urls,
 
 	/* find a valid cache dir to download to */
 	cachedir = _alpm_filecache_setup(handle);
+	temporary_cachedir = _alpm_temporary_download_dir_setup(cachedir, handle->sandboxuser);
+	ASSERT(temporary_cachedir != NULL, return -1);
 
 	for(i = urls; i; i = i->next) {
 		char *url = i->data;
@@ -1034,16 +1319,34 @@ int SYMEXPORT alpm_fetch_pkgurl(alpm_handle_t *handle, const alpm_list_t *urls,
 			CALLOC(payload, 1, sizeof(*payload), GOTO_ERR(handle, ALPM_ERR_MEMORY, err));
 			STRDUP(payload->fileurl, url, FREE(payload); GOTO_ERR(handle, ALPM_ERR_MEMORY, err));
 
+			STRDUP(payload->remote_name, get_filename(payload->fileurl),
+				GOTO_ERR(handle, ALPM_ERR_MEMORY, err));
+
 			c = strrchr(url, '/');
-			if(strstr(c, ".pkg")) {
+			if(c != NULL &&  strstr(c, ".pkg") && payload->remote_name && strlen(payload->remote_name) > 0) {
 				/* we probably have a usable package filename to download to */
+				payload->destfile_name = _alpm_get_fullpath(temporary_cachedir, payload->remote_name, "");
+				payload->tempfile_name = _alpm_get_fullpath(temporary_cachedir, payload->remote_name, ".part");
 				payload->allow_resume = 1;
+
+				if(!payload->destfile_name || !payload->tempfile_name) {
+					goto err;
+				}
+
 			} else {
-				payload->random_partfile = 1;
+				/* The URL does not contain a filename, so download to a temporary location.
+				 * We can not support resuming this kind of download; any partial transfers
+				 * will be destroyed */
+				payload->unlink_on_fail = 1;
+
+				payload->tempfile_openmode = "wb";
+				payload->localf = create_tempfile(payload, temporary_cachedir);
+				if(payload->localf == NULL) {
+					goto err;
+				}
 			}
 
 			payload->handle = handle;
-			payload->trust_remote_name = 1;
 			payload->download_signature = (handle->siglevel & ALPM_SIG_PACKAGE);
 			payload->signature_optional = (handle->siglevel & ALPM_SIG_PACKAGE_OPTIONAL);
 			payloads = alpm_list_add(payloads, payload);
@@ -1054,7 +1357,7 @@ int SYMEXPORT alpm_fetch_pkgurl(alpm_handle_t *handle, const alpm_list_t *urls,
 		event.type = ALPM_EVENT_PKG_RETRIEVE_START;
 		event.pkg_retrieve.num = alpm_list_count(payloads);
 		EVENT(handle, &event);
-		if(_alpm_download(handle, payloads, cachedir) == -1) {
+		if(_alpm_download(handle, payloads, cachedir, temporary_cachedir) == -1) {
 			_alpm_log(handle, ALPM_LOG_WARNING, _("failed to retrieve some files\n"));
 			event.type = ALPM_EVENT_PKG_RETRIEVE_FAILED;
 			EVENT(handle, &event);
@@ -1087,10 +1390,14 @@ int SYMEXPORT alpm_fetch_pkgurl(alpm_handle_t *handle, const alpm_list_t *urls,
 		FREELIST(payloads);
 	}
 
+	_alpm_remove_temporary_download_dir(temporary_cachedir);
+	FREE(temporary_cachedir);
 	return 0;
 
 err:
 	alpm_list_free_inner(payloads, (alpm_list_fn_free)_alpm_dload_payload_reset);
+	_alpm_remove_temporary_download_dir(temporary_cachedir);
+	FREE(temporary_cachedir);
 	FREELIST(payloads);
 	FREELIST(*fetched);
 
@@ -1104,7 +1411,6 @@ void _alpm_dload_payload_reset(struct dload_payload *payload)
 	FREE(payload->remote_name);
 	FREE(payload->tempfile_name);
 	FREE(payload->destfile_name);
-	FREE(payload->content_disp_name);
 	FREE(payload->fileurl);
 	FREE(payload->filepath);
 	*payload = (struct dload_payload){0};
