@@ -1,7 +1,7 @@
 /*
  *  remove.c
  *
- *  Copyright (c) 2006-2021 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2006-2024 Pacman Development Team <pacman-dev@lists.archlinux.org>
  *  Copyright (c) 2002-2006 by Judd Vinet <jvinet@zeroflux.org>
  *  Copyright (c) 2005 by Aurelien Foret <orelien@chez.com>
  *  Copyright (c) 2005 by Christian Hamar <krics@linuxforum.hu>
@@ -420,7 +420,9 @@ static void shift_pacsave(alpm_handle_t *handle, const char *file)
 
 cleanup:
 	free(dirname);
-	closedir(dir);
+	if(dir != NULL) {
+		closedir(dir);
+	}
 }
 
 
@@ -449,10 +451,11 @@ static int unlink_file(alpm_handle_t *handle, alpm_pkg_t *oldpkg,
 		_alpm_log(handle, ALPM_LOG_DEBUG, "path too long to unlink %s%s\n",
 				handle->root, fileobj->name);
 		return -1;
-	} else if(file[file_len-1] == '/') {
+	} else if(file[file_len - 1] == '/') {
 		/* trailing slashes cause errors and confusing messages if the user has
 		 * replaced a directory with a symlink */
-		file[--file_len] = '\0';
+		file[file_len - 1] = '\0';
+		file_len--;
 	}
 
 	if(llstat(file, &buf)) {
@@ -461,9 +464,22 @@ static int unlink_file(alpm_handle_t *handle, alpm_pkg_t *oldpkg,
 	}
 
 	if(S_ISDIR(buf.st_mode)) {
-		ssize_t files = _alpm_files_in_directory(handle, file, 0);
-		/* if we have files, no need to remove the directory */
+		ssize_t files;
+
+		/* restore/add trailing slash */
+		if(file_len < PATH_MAX - 1) {
+			file[file_len] = '/';
+			file_len++;
+			file[file_len] = '\0';
+		} else {
+			_alpm_log(handle, ALPM_LOG_DEBUG, "path too long to unlink %s%s\n",
+					handle->root, fileobj->name);
+			return -1;
+		}
+
+		files = _alpm_files_in_directory(handle, file, 0);
 		if(files > 0) {
+			/* if we have files, no need to remove the directory */
 			_alpm_log(handle, ALPM_LOG_DEBUG, "keeping directory %s (contains files)\n",
 					file);
 		} else if(files < 0) {
