@@ -1,7 +1,7 @@
 /*
  *  signing.c
  *
- *  Copyright (c) 2008-2021 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2008-2024 Pacman Development Team <pacman-dev@lists.archlinux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -316,7 +316,7 @@ static int key_search_keyserver(alpm_handle_t *handle, const char *fpr,
 	 * key fingerprint with 0x, or the lookup will fail. */
 	fpr_len = strlen(fpr);
 	MALLOC(full_fpr, fpr_len + 3, RET_ERR(handle, ALPM_ERR_MEMORY, -1));
-	sprintf(full_fpr, "0x%s", fpr);
+	snprintf(full_fpr, fpr_len + 3, "0x%s", fpr);
 
 	gpg_err = gpgme_new(&ctx);
 	CHECK_ERR();
@@ -356,7 +356,7 @@ static int key_search_keyserver(alpm_handle_t *handle, const char *fpr,
 	pgpkey->data = key;
 	if(key->subkeys->fpr) {
 		pgpkey->fingerprint = key->subkeys->fpr;
-	} else if(key->subkeys->keyid) {
+	} else {
 		pgpkey->fingerprint = key->subkeys->keyid;
 	}
 
@@ -514,19 +514,15 @@ int _alpm_key_import(alpm_handle_t *handle, const char *uid, const char *fpr)
 		return -1;
 	}
 
-	STRDUP(fetch_key.uid, uid, return -1);
-	STRDUP(fetch_key.fingerprint, fpr, free(fetch_key.uid); return -1);
 
 	alpm_question_import_key_t question = {
 				.type = ALPM_QUESTION_IMPORT_KEY,
 				.import = 0,
-				.key = &fetch_key
+				.uid = uid,
+				.fingerprint = fpr
 			};
 	QUESTION(handle, &question);
 
-	free(fetch_key.uid);
-	free(fetch_key.fingerprint);
-
 	if(question.import) {
 		/* Try to import the key from a WKD first */
 		if(email_from_uid(uid, &email) == 0) {
@@ -820,7 +816,7 @@ char *_alpm_sigpath(alpm_handle_t *handle, const char *path)
 	}
 	len = strlen(path) + 5;
 	CALLOC(sigpath, len, sizeof(char), RET_ERR(handle, ALPM_ERR_MEMORY, NULL));
-	sprintf(sigpath, "%s.sig", path);
+	snprintf(sigpath, len, "%s.sig", path);
 	return sigpath;
 }
 
@@ -1048,7 +1044,7 @@ int SYMEXPORT alpm_siglist_cleanup(alpm_siglist_t *siglist)
 static size_t length_check(size_t length, size_t position, size_t a,
 		alpm_handle_t *handle, const char *identifier)
 {
-	if( a == 0 || length - position <= a) {
+	if( a == 0 || position > length || length - position <= a) {
 		_alpm_log(handle, ALPM_LOG_ERROR,
 				_("%s: signature format error\n"), identifier);
 		return -1;
@@ -1089,7 +1085,7 @@ static int parse_subpacket(alpm_handle_t *handle, const char *identifier,
 					return -1;
 				}
 				for (i = 0; i < 8; i++) {
-					sprintf(&key[i * 2], "%02X", sig[spos + i + 1]);
+					snprintf(&key[i * 2], 3, "%02X", sig[spos + i + 1]);
 				}
 				*keys = alpm_list_add(*keys, strdup(key));
 				break;
